/*---------------------------------------------------------
 * Copyright (C) Microsoft Corporation. All rights reserved.
 *--------------------------------------------------------*/
'use strict';
var languageServerError_1 = require('./languageServerError');
exports.LanguageServerError = languageServerError_1.LanguageServerError;
exports.MessageKind = languageServerError_1.MessageKind;
var vscode_jsonrpc_1 = require('vscode-jsonrpc');
exports.ResponseError = vscode_jsonrpc_1.ResponseError;
exports.ErrorCodes = vscode_jsonrpc_1.ErrorCodes;
var protocol_1 = require('./protocol');
exports.FileChangeType = protocol_1.FileChangeType;
exports.Severity = protocol_1.Severity;
var textDocuments_1 = require('./textDocuments');
// ------------- Reexport the API surface of the language worker API ----------------------
var fm = require('./files');
var Files;
(function (Files) {
    Files.uriToFilePath = fm.uriToFilePath;
    Files.resolveModule = fm.resolveModule;
})(Files = exports.Files || (exports.Files = {}));
// ------------------------- implementation of the language worker protocol ---------------------------------------------
function isFunction(arg) {
    return Object.prototype.toString.call(arg) === '[object Function]';
}
function isArray(array) {
    if (Array.isArray) {
        return Array.isArray(array);
    }
    if (array && typeof (array.length) === 'number' && array.constructor === Array) {
        return true;
    }
    return false;
}
var DocumentManager = (function () {
    function DocumentManager() {
        this.trackedDocuments = Object.create(null);
    }
    DocumentManager.prototype.add = function (uri, document) {
        this.trackedDocuments[uri] = document;
    };
    DocumentManager.prototype.remove = function (uri) {
        return delete this.trackedDocuments[uri];
    };
    DocumentManager.prototype.get = function (uri) {
        return this.trackedDocuments[uri];
    };
    DocumentManager.prototype.all = function () {
        var _this = this;
        return Object.keys(this.trackedDocuments).map(function (key) { return _this.trackedDocuments[key]; });
    };
    DocumentManager.prototype.keys = function () {
        return Object.keys(this.trackedDocuments);
    };
    return DocumentManager;
})();
var ErrorMessageTracker = (function () {
    function ErrorMessageTracker() {
        this.messages = Object.create(null);
    }
    ErrorMessageTracker.prototype.add = function (message) {
        var count = this.messages[message];
        if (!count) {
            count = 0;
        }
        count++;
        this.messages[message] = count;
    };
    ErrorMessageTracker.prototype.publish = function (connection) {
        Object.keys(this.messages).forEach(function (message) {
            connection.window.showErrorMessage(message);
        });
    };
    return ErrorMessageTracker;
})();
var Logger = (function () {
    function Logger() {
    }
    Logger.prototype.attach = function (connection) {
        this.connection = connection;
    };
    Logger.prototype.error = function (message) {
        this.send(protocol_1.MessageType.Error, message);
    };
    Logger.prototype.warn = function (message) {
        this.send(protocol_1.MessageType.Warning, message);
    };
    Logger.prototype.info = function (message) {
        this.send(protocol_1.MessageType.Info, message);
    };
    Logger.prototype.log = function (message) {
        this.send(protocol_1.MessageType.Log, message);
    };
    Logger.prototype.send = function (type, message) {
        if (this.connection) {
            this.connection.sendNotification(protocol_1.LogMessageNotification.type, { type: type, message: message });
        }
    };
    return Logger;
})();
var RemoteWindowImpl = (function () {
    function RemoteWindowImpl(connection) {
        this.connection = connection;
    }
    RemoteWindowImpl.prototype.showErrorMessage = function (message) {
        this.connection.sendNotification(protocol_1.ShowMessageNotification.type, { type: protocol_1.MessageType.Error, message: message });
    };
    RemoteWindowImpl.prototype.showWarningMessage = function (message) {
        this.connection.sendNotification(protocol_1.ShowMessageNotification.type, { type: protocol_1.MessageType.Warning, message: message });
    };
    RemoteWindowImpl.prototype.showInformationMessage = function (message) {
        this.connection.sendNotification(protocol_1.ShowMessageNotification.type, { type: protocol_1.MessageType.Info, message: message });
    };
    return RemoteWindowImpl;
})();
function createConnection(inputStream, outputStream) {
    var shutdownReceived;
    inputStream.on('end', function () {
        process.exit(shutdownReceived ? 0 : 1);
    });
    inputStream.on('close', function () {
        process.exit(shutdownReceived ? 0 : 1);
    });
    var logger = new Logger();
    var connection = vscode_jsonrpc_1.createServerMessageConnection(inputStream, outputStream, logger);
    logger.attach(connection);
    var remoteWindow = new RemoteWindowImpl(connection);
    var shutdownHandler = null;
    connection.onRequest(protocol_1.ShutdownRequest.type, function (params) {
        shutdownReceived = true;
        if (shutdownHandler) {
            return shutdownHandler(params);
        }
        else {
            return undefined;
        }
    });
    var result = {
        onRequest: function (type, handler) { return connection.onRequest(type, handler); },
        sendNotification: function (type, params) { return connection.sendNotification(type, params); },
        onNotification: function (type, handler) { return connection.onNotification(type, handler); },
        onInitialize: function (handler) { return connection.onRequest(protocol_1.InitializeRequest.type, handler); },
        onShutdown: function (handler) { return shutdownHandler = handler; },
        onExit: function (handler) { return connection.onNotification(protocol_1.ExitNotification.type, handler); },
        get console() { return logger; },
        get window() { return remoteWindow; },
        onDidChangeConfiguration: function (handler) { return connection.onNotification(protocol_1.DidChangeConfigurationNotification.type, handler); },
        onDidChangeFiles: function (handler) { return connection.onNotification(protocol_1.DidChangeFilesNotification.type, handler); },
        onDidOpenTextDocument: function (handler) { return connection.onNotification(protocol_1.DidOpenTextDocumentNotification.type, handler); },
        onDidChangeTextDocument: function (handler) { return connection.onNotification(protocol_1.DidChangeTextDocumentNotification.type, handler); },
        onDidCloseTextDocument: function (handler) { return connection.onNotification(protocol_1.DidCloseTextDocumentNotification.type, handler); },
        publishDiagnostics: function (params) { return connection.sendNotification(protocol_1.PublishDiagnosticsNotification.type, params); },
        dispose: function () { return connection.dispose(); }
    };
    return result;
}
exports.createConnection = createConnection;
var ValidationRequestor = (function () {
    function ValidationRequestor(documents) {
        this.documents = documents;
        this._toValidate = Object.create(null);
    }
    Object.defineProperty(ValidationRequestor.prototype, "toValidate", {
        get: function () {
            var _this = this;
            return Object.keys(this._toValidate).map(function (key) { return _this._toValidate[key]; });
        },
        enumerable: true,
        configurable: true
    });
    ValidationRequestor.prototype.all = function () {
        var _this = this;
        this.documents.keys().forEach(function (key) { return _this._toValidate[key] = _this.documents.get(key); });
    };
    ValidationRequestor.prototype.single = function (uri) {
        var document = this.documents.get(uri);
        if (document) {
            this._toValidate[uri] = document;
        }
        return !!document;
    };
    return ValidationRequestor;
})();
function createValidatorConnection(inputStream, outputStream) {
    var connection = createConnection(inputStream, outputStream);
    var documents = new DocumentManager();
    function run(handler) {
        var rootFolder;
        function doProcess(result, complete, error) {
            if (isFunction(result.then)) {
                return result.then(complete, error);
            }
            else {
                return complete(result);
            }
        }
        function safeRunner(values, func) {
            var messageTracker = new ErrorMessageTracker();
            var runSingle = function (value) {
                try {
                    func(value);
                }
                catch (error) {
                    if (error instanceof languageServerError_1.LanguageServerError) {
                        var workerError = error;
                        switch (workerError.messageKind) {
                            case languageServerError_1.MessageKind.Show:
                                messageTracker.add(workerError.message);
                                break;
                            case languageServerError_1.MessageKind.Log:
                                logSafeRunnerMessage(workerError.message);
                                break;
                        }
                    }
                    else {
                        logSafeRunnerMessage(error.message);
                    }
                }
            };
            if (isArray(values)) {
                for (var _i = 0, _a = values; _i < _a.length; _i++) {
                    var value = _a[_i];
                    runSingle(value);
                }
            }
            else {
                runSingle(values);
            }
            messageTracker.publish(connection);
        }
        function logSafeRunnerMessage(message) {
            if (message) {
                connection.console.error("Safe Runner failed with message: " + message);
            }
            else {
                connection.console.error('Safe Runner failed unexpectedly.');
            }
        }
        function validate(document) {
            var result = handler.validate(document);
            doProcess(result, function (diagnostics) {
                connection.publishDiagnostics({
                    uri: document.uri,
                    diagnostics: diagnostics
                });
            }, function (error) {
                // We need a log event to tell the client that a diagnostic
                // failed.
            });
        }
        function createInitializeResult(initArgs) {
            var resultCapabilities = {
                incrementalTextDocumentSync: false
            };
            return { capabilities: resultCapabilities };
        }
        connection.onInitialize(function (initArgs) {
            rootFolder = initArgs.rootFolder;
            if (isFunction(handler.initialize)) {
                return doProcess(handler.initialize(rootFolder), function (resultOrError) {
                    if (resultOrError) {
                        return resultOrError;
                    }
                    else {
                        return createInitializeResult(initArgs);
                    }
                });
            }
            else {
                return createInitializeResult(initArgs);
            }
        });
        connection.onShutdown(function (params) {
            if (isFunction(handler.shutdown)) {
                handler.shutdown();
            }
            return undefined;
        });
        connection.onExit(function () {
            process.exit(0);
        });
        connection.onDidOpenTextDocument(function (event) {
            var document = new textDocuments_1.SimpleTextDocument(event.uri, event.text);
            documents.add(event.uri, document);
            process.nextTick(function () { return safeRunner(document, validate); });
        });
        connection.onDidChangeTextDocument(function (event) {
            var document = documents.get(event.uri);
            if (document) {
                document.setText(event.text);
                process.nextTick(function () { return safeRunner(document, validate); });
            }
        });
        connection.onDidCloseTextDocument(function (event) {
            documents.remove(event.uri);
        });
        connection.onDidChangeConfiguration(function (eventBody) {
            var settings = eventBody.settings;
            var requestor = new ValidationRequestor(documents);
            if (isFunction(handler.onConfigurationChange)) {
                handler.onConfigurationChange(settings, requestor);
            }
            else {
                requestor.all();
            }
            process.nextTick(function () { return safeRunner(requestor.toValidate, validate); });
        });
        connection.onDidChangeFiles(function (args) {
            var requestor = new ValidationRequestor(documents);
            if (isFunction(handler.onFileEvents)) {
                handler.onFileEvents(args.changes, requestor);
            }
            else {
                requestor.all();
            }
            process.nextTick(function () { return safeRunner(requestor.toValidate, validate); });
        });
    }
    var result = {
        onRequest: function (type, handler) { return connection.onRequest(type, handler); },
        sendNotification: function (type, params) { return connection.sendNotification(type, params); },
        onNotification: function (type, handler) { return connection.onNotification(type, handler); },
        get console() { return connection.console; },
        get window() { return connection.window; },
        run: run
    };
    return result;
}
exports.createValidatorConnection = createValidatorConnection;
function runSingleFileValidator(inputStream, outputStream, handler) {
    createValidatorConnection(inputStream, outputStream).run(handler);
}
exports.runSingleFileValidator = runSingleFileValidator;
